from __future__ import annotations

import json
import os
import platform
import sys
from functools import lru_cache
from typing import Any

from hatch._version import __version__


def get_linehaul_data() -> dict[str, Any]:
    data: dict[str, Any] = {
        "installer": {"name": "hatch", "version": __version__},
        "python": platform.python_version(),
        "implementation": {
            "name": platform.python_implementation(),
            "version": _get_implementation_version(),
        },
        "system": {
            "name": platform.system(),
            "release": platform.release(),
        },
        "cpu": platform.machine(),
        "ci": _looks_like_ci() or None,
    }

    if sys.platform.startswith("linux"):
        distro_info = _get_linux_distro_info()
        if distro_info:
            data["distro"] = distro_info
    elif sys.platform == "darwin":
        mac_version = platform.mac_ver()[0]
        if mac_version:
            data["distro"] = {"name": "macOS", "version": mac_version}

    openssl_version = _get_openssl_version()
    if openssl_version:
        data["openssl_version"] = openssl_version

    return data


def _get_implementation_version() -> str:
    implementation_name = platform.python_implementation()

    if implementation_name == "PyPy":
        pypy_version_info = sys.pypy_version_info  # type: ignore[attr-defined]
        if pypy_version_info.releaselevel == "final":
            pypy_version_info = pypy_version_info[:3]
        return ".".join(str(x) for x in pypy_version_info)

    return platform.python_version()


def _get_linux_distro_info() -> dict[str, Any] | None:
    import distro

    name = distro.name()
    if not name:
        return None

    distro_info: dict[str, Any] = {"name": name}

    version = distro.version()
    if version:
        distro_info["version"] = version

    codename = distro.codename()
    if codename:
        distro_info["id"] = codename

    libc_info = _get_libc_info()
    if libc_info:
        distro_info["libc"] = libc_info

    return distro_info


def _get_libc_info() -> dict[str, str] | None:
    libc, version = _get_libc_version()
    if libc and version:
        return {"lib": libc, "version": version}
    return None


def _get_libc_version() -> tuple[str | None, str | None]:
    import ctypes
    import ctypes.util

    libc_name = ctypes.util.find_library("c")
    if not libc_name:
        return None, None

    try:
        libc = ctypes.CDLL(libc_name)
        libc.gnu_get_libc_version.restype = ctypes.c_char_p
        version = libc.gnu_get_libc_version()
        if version:
            return "glibc", version.decode()
    except (OSError, AttributeError, UnicodeDecodeError):
        pass

    return None, None


def _get_openssl_version() -> str | None:
    try:
        import ssl
    except (ImportError, AttributeError):
        return None
    else:
        return ssl.OPENSSL_VERSION


def _looks_like_ci() -> bool:
    # INFO: Matches pip's CI detection for linehaul statistics.
    # Uses existence check (not value check) because variables like BUILD_BUILDID
    # and BUILD_ID are set to build IDs, not boolean strings.
    ci_env_vars = (
        "BUILD_BUILDID",
        "BUILD_ID",
        "CI",
        "GITHUB_ACTIONS",
        "GITLAB_CI",
        "PIP_IS_CI",
        "TRAVIS",
    )
    return any(env_var in os.environ for env_var in ci_env_vars)


@lru_cache(maxsize=1)
def get_linehaul_component() -> str:
    data = get_linehaul_data()
    return json.dumps(data, separators=(",", ":"), sort_keys=True)



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import click

from hatch.config.constants import AppEnvVars

if TYPE_CHECKING:
    from hatch.cli.application import Application


def filter_environments(environments, filter_data):
    selected_environments = []
    for env_name, env_data in environments.items():
        for key, value in filter_data.items():
            if key not in env_data or env_data[key] != value:
                break
        else:
            selected_environments.append(env_name)

    return selected_environments


@click.command(short_help="Run commands within project environments")
@click.argument("args", required=True, nargs=-1)
@click.option("--env", "-e", "env_names", multiple=True, help="The environments to target")
@click.option("--include", "-i", "included_variable_specs", multiple=True, help="The matrix variables to include")
@click.option("--exclude", "-x", "excluded_variable_specs", multiple=True, help="The matrix variables to exclude")
@click.option("--filter", "-f", "filter_json", default=None, help="The JSON data used to select environments")
@click.option(
    "--force-continue", is_flag=True, help="Run every command and if there were any errors exit with the first code"
)
@click.option("--ignore-compat", is_flag=True, help="Ignore incompatibility when selecting specific environments")
@click.pass_obj
def run(
    app: Application,
    *,
    args: tuple[str, ...],
    env_names: tuple[str, ...],
    included_variable_specs: tuple[str, ...],
    excluded_variable_specs: tuple[str, ...],
    filter_json: str | None,
    force_continue: bool,
    ignore_compat: bool,
):
    """
    Run commands within project environments.

    The `-e`/`--env` option overrides the equivalent [root option](#hatch) and the `HATCH_ENV` environment variable.

    The `-i`/`--include` and `-x`/`--exclude` options may be used to include or exclude certain
    variables, optionally followed by specific comma-separated values, and may be selected multiple
    times. For example, if you have the following configuration:

    \b
    ```toml config-example
    [[tool.hatch.envs.test.matrix]]
    python = ["3.9", "3.10"]
    version = ["42", "3.14", "9000"]
    ```

    then running:

    \b
    ```
    hatch env run -i py=3.10 -x version=9000 test:pytest
    ```

    would execute `pytest` in the environments `test.py3.10-42` and `test.py3.10-3.14`.
    Note that `py` may be used as an alias for `python`.

    \b
    !!! note
        The inclusion option is treated as an intersection while the exclusion option is treated as a
        union i.e. an environment must match all of the included variables to be selected while matching
        any of the excluded variables will prevent selection.
    """
    from hatch.utils.runner import parse_matrix_variables, select_environments

    try:
        included_variables = parse_matrix_variables(included_variable_specs)
    except ValueError as e:
        app.abort(f"Duplicate included variable: {e}")

    try:
        excluded_variables = parse_matrix_variables(excluded_variable_specs)
    except ValueError as e:
        app.abort(f"Duplicate excluded variable: {e}")

    app.ensure_environment_plugin_dependencies()

    project = app.project

    if not env_names:
        env_names = (app.env,)
    elif "system" in env_names:
        project.config.config["envs"] = {
            "system": {
                "type": "system",
                "skip-install": True,
                "scripts": project.config.scripts,
            }
        }

    # Deduplicate
    ordered_env_names = list(dict.fromkeys(env_names))

    environments = []
    matrix_selected = False
    for env_name in ordered_env_names:
        if env_name in project.config.matrices:
            matrix_selected = True
            env_data = project.config.matrices[env_name]["envs"]
            if not env_data:
                app.abort(f"No variables defined for matrix: {env_name}")

            environments.extend(select_environments(env_data, included_variables, excluded_variables))
        else:
            environments.append(env_name)

    if filter_json:
        import json

        filter_data = json.loads(filter_json)
        if not isinstance(filter_data, dict):
            app.abort("The --filter/-f option must be a JSON mapping")

        environments[:] = filter_environments(project.config.envs, filter_data)

    if not environments:
        app.abort("No environments were selected")
    elif not matrix_selected and (included_variables or excluded_variables):
        app.abort(f"Variable selection is unsupported for non-matrix environments: {', '.join(ordered_env_names)}")

    for context in app.runner_context(
        environments,
        ignore_compat=ignore_compat or matrix_selected,
        display_header=matrix_selected,
        keep_env=bool(os.environ.get(AppEnvVars.KEEP_ENV)),
    ):
        if context.env.name == "system":
            context.env.exists = lambda: True  # type: ignore[method-assign]

        context.force_continue = force_continue
        context.add_shell_command(list(args))



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Verify formatting", context_settings={"ignore_unknown_options": True})
@click.argument("args", nargs=-1)
@click.option("--fix", is_flag=True, help="Apply formatting fixes rather than just checking")
@click.option("--sync", is_flag=True, help="Sync the default config file with the current version of Hatch")
@click.pass_obj
def fmt(
    app: Application,
    *,
    args: tuple[str, ...],
    fix: bool,
    sync: bool,
):
    """
    Verify formatting, using Ruff by default.
    """
    from hatch.cli.fmt.core import StaticAnalysisEnvironment

    app.ensure_environment_plugin_dependencies()

    for context in app.runner_context(["hatch-check-fmt"]):
        sa_env = StaticAnalysisEnvironment(context.env)

        # TODO: remove in a few minor releases, this is very new but we don't want to break users on the cutting edge
        if legacy_config_path := app.project.config.config.get("format", {}).get("config-path", ""):
            app.display_warning(
                "The `tool.hatch.format.config-path` option is deprecated and will be removed in a future release. "
                "Use `tool.hatch.envs.hatch-check-fmt.config-path` instead."
            )
            sa_env.config_path = legacy_config_path

        if sync and not sa_env.config_path:
            app.abort("The --sync flag can only be used when the `tool.hatch.format.config-path` option is defined")

        script = "format-fix" if fix else "format-check"

        default_args = sa_env.get_default_args()
        arguments = list(args)
        try:
            arguments.remove("--preview")
        except ValueError:
            preview = False
        else:
            preview = True
            default_args.append("--preview")

        internal_args = context.env.join_command_args(default_args)
        if internal_args:
            internal_args = f" {internal_args}"

        formatted_args = context.env.join_command_args(arguments)
        context.add_shell_command(f"{script} {formatted_args}")

        # Check for legacy HATCH_FMT_ARGS usage in scripts and warn about deprecation
        for script_body in context.env.config.get("scripts", {}).values():
            if isinstance(script_body, str) and "HATCH_FMT_ARGS" in script_body:
                app.display_warning(
                    "The `HATCH_FMT_ARGS` environment variable is deprecated. Use `HATCH_CHECK_FMT_ARGS` instead."
                )
                break

        context.env_vars["HATCH_CHECK_FMT_ARGS"] = internal_args

        if not sa_env.config_path or sync:
            sa_env.write_config_file(preview=preview)



import click


@click.command(short_help="Create or initialize a project")
@click.argument("name", required=False)
@click.argument("location", required=False)
@click.option("--interactive", "-i", "interactive", is_flag=True, help="Interactively choose details about the project")
@click.option("--cli", "feature_cli", is_flag=True, help="Give the project a command line interface")
@click.option("--init", "initialize", is_flag=True, help="Initialize an existing project")
@click.option("-so", "setuptools_options", multiple=True, hidden=True)
@click.pass_obj
def new(app, name, location, interactive, feature_cli, initialize, setuptools_options):
    """Create or initialize a project."""
    import sys
    from copy import deepcopy
    from datetime import datetime, timezone

    from hatch.config.model import TemplateConfig
    from hatch.project.core import Project
    from hatch.template import File
    from hatch.utils.fs import Path

    if location:
        location = Path(location).resolve()

    migration_possible = False
    if initialize:
        interactive = True
        location = location or Path.cwd()
        if (location / "setup.py").is_file() or (location / "setup.cfg").is_file():
            migration_possible = True
            if not name:
                name = "temporary"

    if not name:
        if not interactive:
            app.abort("Missing required argument for the project name, use the -i/--interactive flag.")

        name = app.prompt("Project name")

    normalized_name = Project.canonicalize_name(name, strict=False)
    if not location:
        location = Path(normalized_name).resolve()

    needs_config_update = False
    if location.is_file():
        app.abort(f"Path `{location}` points to a file.")
    elif location.is_dir() and any(location.iterdir()):
        if not initialize:
            app.abort(f"Directory `{location}` is not empty.")

        needs_config_update = (location / "pyproject.toml").is_file()

    if migration_possible:
        from hatch.cli.new.migrate import migrate
        from hatch.venv.core import TempVirtualEnv

        try:
            with (
                app.status("Migrating project metadata from setuptools"),
                TempVirtualEnv(sys.executable, app.platform) as venv,
            ):
                app.platform.run_command(["python", "-m", "pip", "install", "-q", "setuptools"])
                migrate(str(location), setuptools_options, venv.sys_path)
        except Exception as e:  # noqa: BLE001
            app.display_error(f"Could not automatically migrate from setuptools: {e}")
            if name == "temporary":
                name = app.prompt("Project name")
        else:
            return

    default_config = {
        "description": "",
        "dependencies": set(),
        "package_name": normalized_name.replace("-", "_"),
        "project_name": name,
        "project_name_normalized": normalized_name,
        "args": {"cli": feature_cli},
    }

    if interactive:
        default_config["description"] = app.prompt("Description", default="")

        app.display_info()

    if needs_config_update:
        app.project.initialize(str(location / "pyproject.toml"), default_config)
        app.display_success("Updated: pyproject.toml")
        return

    template_config = deepcopy(app.config.template.raw_data)
    if "plugins" in template_config and not template_config["plugins"]:
        del template_config["plugins"]

    TemplateConfig(template_config, ("template",)).parse_fields()

    plugin_config = template_config.pop("plugins")

    # Set up default config for template files
    template_config.update(default_config)

    template_classes = app.plugins.template.collect()

    templates = []
    for template_name, template_class in sorted(template_classes.items(), key=lambda item: -item[1].PRIORITY):
        if template_name in plugin_config:
            templates.append(
                template_class(plugin_config.pop(template_name), app.cache_dir, datetime.now(timezone.utc))
            )

    if not templates:
        app.abort(f"None of the defined plugins were found: {', '.join(sorted(plugin_config))}")
    elif plugin_config:
        app.abort(f"Some of the defined plugins were not found: {', '.join(sorted(plugin_config))}")

    for template in templates:
        template.initialize_config(template_config)

    template_files = []

    for template in templates:
        for possible_template_file in template.get_files(config=deepcopy(template_config)):
            template_file = (
                possible_template_file(deepcopy(template_config), template.plugin_config)
                if possible_template_file.__class__ is not File
                else possible_template_file
            )
            if template_file.path is None or (initialize and str(template_file.path) != "pyproject.toml"):
                continue

            template_files.append(template_file)

    for template in templates:
        template.finalize_files(config=deepcopy(template_config), files=template_files)

    for template_file in template_files:
        template_file.write(location)

    if initialize:
        app.display_success("Wrote: pyproject.toml")
        return

    from rich.markup import escape
    from rich.tree import Tree

    def recurse_directory(directory, tree):
        paths = sorted(Path(directory).iterdir(), key=lambda p: (p.is_file(), p.name))
        for path in paths:
            if path.is_dir():
                recurse_directory(
                    path, tree.add(f"[bold magenta][link={app.platform.format_file_uri(path)}]{escape(path.name)}")
                )
            else:
                tree.add(f"[green][link={app.platform.format_file_uri(path)}]{escape(path.name)}")

    root = Tree(
        f"[bold magenta][link={app.platform.format_file_uri(location)}]{escape(location.name)}",
        guide_style="bright_blue",
        hide_root=location == Path.cwd(),
    )
    recurse_directory(location, root)
    app.output(root, markup=True)



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Display project metadata")
@click.argument("field", required=False)
@click.pass_obj
def metadata(app: Application, field: str | None):
    """
    Display project metadata.

    If you want to view the raw readme file without rendering, you can use a JSON parser
    like [jq](https://github.com/stedolan/jq):

    \b
    ```
    hatch project metadata | jq -r .readme
    ```
    """
    app.ensure_environment_plugin_dependencies()

    import json

    from hatch.project.constants import BUILD_BACKEND

    app.project.prepare_build_environment()
    build_backend = app.project.metadata.build.build_backend
    with app.project.location.as_cwd(), app.project.build_env.get_env_vars():
        if build_backend != BUILD_BACKEND:
            project_metadata = app.project.build_frontend.get_core_metadata()
        else:
            project_metadata = app.project.build_frontend.hatch.get_core_metadata()

    if field:
        if field not in project_metadata:
            app.abort(f"Unknown metadata field: {field}")
        elif field == "readme":
            if project_metadata[field]["content-type"] == "text/markdown":  # no cov
                app.display_markdown(project_metadata[field]["text"])
            else:
                app.display(project_metadata[field]["text"])
        elif isinstance(project_metadata[field], str):
            app.display(project_metadata[field])
        else:
            app.display(json.dumps(project_metadata[field], indent=4))
    else:
        for key, value in list(project_metadata.items()):
            if not value:
                project_metadata.pop(key)

        app.display(json.dumps(project_metadata, indent=4))



from __future__ import annotations

from typing import Any


def get_default_config() -> dict[str, Any]:
    return {
        "installer": "uv",
        "dependencies": [
            "coverage-enable-subprocess==1.0",
            "coverage[toml]~=7.11",
            "pytest~=9.0",
            "pytest-mock~=3.12",
            "pytest-randomly~=3.15",
            "pytest-rerunfailures~=14.0",
            "pytest-xdist[psutil]~=3.5",
        ],
        "scripts": {
            "run": "pytest{env:HATCH_TEST_ARGS:} {args}",
            "run-cov": "coverage run -m pytest{env:HATCH_TEST_ARGS:} {args}",
            "cov-combine": "coverage combine",
            "cov-report": "coverage report",
        },
        "matrix": [{"python": ["3.14", "3.14t", "3.13", "3.12", "3.11", "3.10"]}],
    }



from __future__ import annotations

import os

from hatch.config.constants import AppEnvVars


def get_env_var(*, plugin_name: str, option: str) -> str:
    return f"{AppEnvVars.ENV_OPTION_PREFIX}{plugin_name}_{option.replace('-', '_')}".upper()


def get_env_var_option(*, plugin_name: str, option: str, default: str = "") -> str:
    return os.environ.get(get_env_var(plugin_name=plugin_name, option=option), default)


def ensure_valid_environment(env_config: dict):
    env_config.setdefault("type", "virtual")

    workspace = env_config.get("workspace")
    if workspace is not None:
        if not isinstance(workspace, dict):
            msg = "Field workspace must be a table"
            raise TypeError(msg)

        members = workspace.get("members", [])
        if not isinstance(members, list):
            msg = "Field workspace.members must be an array"
            raise TypeError(msg)

        # Validate each member
        for i, member in enumerate(members, 1):
            if isinstance(member, str):
                continue
            if isinstance(member, dict):
                path = member.get("path")
                if path is None:
                    msg = f"Member #{i} must define a `path` key"
                    raise TypeError(msg)
                if not isinstance(path, str):
                    msg = f"Member #{i} path must be a string"
                    raise TypeError(msg)
            else:
                msg = f"Member #{i} must be a string or table"
                raise TypeError(msg)

        exclude = workspace.get("exclude", [])
        if not isinstance(exclude, list):
            msg = "Field workspace.exclude must be an array"
            raise TypeError(msg)


def get_verbosity_flag(verbosity: int, *, adjustment=0) -> str:
    verbosity += adjustment
    if not verbosity:
        return ""

    if verbosity > 0:
        return f"-{'v' * abs(min(verbosity, 3))}"

    return f"-{'q' * abs(max(verbosity, -3))}"


def add_verbosity_flag(command: list[str], verbosity: int, *, adjustment=0):
    flag = get_verbosity_flag(verbosity, adjustment=adjustment)
    if flag:
        command.append(flag)



from __future__ import annotations

import os
import sys
import sysconfig
from contextlib import contextmanager, nullcontext, suppress
from functools import cached_property
from os.path import isabs
from typing import TYPE_CHECKING

from hatch.config.constants import AppEnvVars
from hatch.env.plugin.interface import EnvironmentInterface
from hatch.env.utils import add_verbosity_flag
from hatch.utils.fs import Path
from hatch.utils.shells import ShellManager
from hatch.utils.structures import EnvVars
from hatch.venv.core import UVVirtualEnv, VirtualEnv

FREETHREADED_BUILD = bool(sysconfig.get_config_var("Py_GIL_DISABLED"))

if TYPE_CHECKING:
    from collections.abc import Callable, Iterable

    from packaging.specifiers import SpecifierSet
    from python_discovery import PythonInfo

    from hatch.dep.core import Dependency
    from hatch.dep.sync import InstalledDistributions
    from hatch.python.core import PythonManager


class VirtualEnvironment(EnvironmentInterface):
    PLUGIN_NAME = "virtual"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        project_id = self.root.id
        project_is_script = self.root.is_file()
        project_name = (
            project_id
            if project_is_script
            else self.metadata.name
            if "project" in self.metadata.config
            else f"{project_id}-unmanaged"
        )
        venv_name = project_name if self.name == "default" else self.name

        # Conditions requiring a flat structure for build env
        if (
            self.isolated_data_directory == self.platform.home / ".virtualenvs"
            or self.root in self.isolated_data_directory.resolve().parents
        ):
            app_virtual_env_path = self.isolated_data_directory / venv_name
        else:
            app_virtual_env_path = self.isolated_data_directory / project_name / project_id / venv_name

        # Explicit path
        chosen_directory = self.get_env_var_option("path") or self.config.get("path", "")
        if chosen_directory:
            self.storage_path = self.data_directory / project_name / project_id
            self.virtual_env_path = (
                Path(chosen_directory) if isabs(chosen_directory) else (self.root / chosen_directory).resolve()
            )
        elif project_is_script:
            self.storage_path = self.virtual_env_path = self.isolated_data_directory / venv_name
        # Conditions requiring a flat structure
        elif (
            self.data_directory == self.platform.home / ".virtualenvs"
            or self.root in self.data_directory.resolve().parents
        ):
            self.storage_path = self.data_directory
            self.virtual_env_path = self.storage_path / venv_name
        # Otherwise the defined app path
        else:
            self.storage_path = self.data_directory / project_name / project_id
            self.virtual_env_path = self.storage_path / venv_name

        self.virtual_env = self.virtual_env_cls(self.virtual_env_path, self.platform, self.verbosity)
        self.build_virtual_env = self.virtual_env_cls(
            app_virtual_env_path.parent / f"{app_virtual_env_path.name}-build", self.platform, self.verbosity
        )
        self.shells = ShellManager(self)

        self._parent_python = None

    @cached_property
    def use_uv(self) -> bool:
        return self.installer == "uv" or bool(self.explicit_uv_path)

    @cached_property
    def installer(self) -> str:
        return self.config.get("installer", "pip")

    @cached_property
    def explicit_uv_path(self) -> str:
        return self.get_env_var_option("uv_path") or self.config.get("uv-path", "")

    @cached_property
    def virtual_env_cls(self) -> type[VirtualEnv]:
        return UVVirtualEnv if self.use_uv else VirtualEnv

    def expose_uv(self):
        if not (self.use_uv or self.uv_path):
            return nullcontext()

        return EnvVars({"HATCH_UV": self.uv_path})

    @cached_property
    def uv_path(self) -> str:
        if self.explicit_uv_path:
            return self.explicit_uv_path

        from hatch.env.internal import is_default_environment

        env_name = "hatch-uv"
        if not (
            # Prevent recursive loop
            self.name == env_name
            # Only if dependencies have been set by the user
            or is_default_environment(env_name, self.app.project.config.internal_envs[env_name])
        ):
            uv_env = self.app.project.get_environment(env_name)
            self.app.project.prepare_environment(uv_env, keep_env=bool(os.environ.get(AppEnvVars.KEEP_ENV)))
            with uv_env:
                return self.platform.modules.shutil.which("uv")

        import sysconfig

        scripts_dir = sysconfig.get_path("scripts")
        old_path = os.environ.get("PATH", os.defpath)
        new_path = f"{scripts_dir}{os.pathsep}{old_path}"
        return self.platform.modules.shutil.which("uv", path=new_path)

    @cached_property
    def distributions(self) -> InstalledDistributions:
        from hatch.dep.sync import InstalledDistributions

        return InstalledDistributions(sys_path=self.virtual_env.sys_path, environment=self.virtual_env.environment)

    @cached_property
    def missing_dependencies(self) -> list[Dependency]:
        return self.distributions.missing_dependencies(self.all_dependencies_complex)

    @staticmethod
    def get_option_types() -> dict:
        return {
            "system-packages": bool,
            "path": str,
            "python-sources": list,
            "installer": str,
            "uv-path": str,
            "locker": str,
        }

    def activate(self):
        self.virtual_env.activate()

    def deactivate(self):
        self.virtual_env.deactivate()

    def find(self):
        return self.virtual_env_path

    def create(self):
        if self.root in self.storage_path.parents:
            # Although it would be nice to support Mercurial, only Git supports multiple ignore files. See:
            # https://github.com/pytest-dev/pytest/issues/3286#issuecomment-421439197
            vcs_ignore_file = self.storage_path / ".gitignore"
            if not vcs_ignore_file.is_file():
                vcs_ignore_file.ensure_parent_dir_exists()
                vcs_ignore_file.write_text(
                    """\
# This file was automatically created by Hatch
*
"""
                )

        with self.expose_uv():
            self.virtual_env.create(self.parent_python, allow_system_packages=self.config.get("system-packages", False))

    def remove(self):
        self.virtual_env.remove()
        self.build_virtual_env.remove()

        # Clean up root directory of all virtual environments belonging to the project
        if self.storage_path != self.platform.home / ".virtualenvs" and self.storage_path.is_dir():
            entries = [entry.name for entry in self.storage_path.iterdir()]
            if not entries or (entries == [".gitignore"] and self.root in self.storage_path.parents):
                self.storage_path.remove()

    def exists(self):
        return self.virtual_env.exists()

    def install_project(self):
        with self.safe_activation():
            self.platform.check_command(self.construct_pip_install_command([self.apply_features(str(self.root))]))

    def install_project_dev_mode(self):
        with self.safe_activation():
            self.platform.check_command(
                self.construct_pip_install_command(["--editable", self.apply_features(str(self.root))])
            )

    def uv_pip_sync_command(self, lockfile_path: Path, *, dry_run: bool = False) -> list[str]:
        command = [self.uv_path, "pip", "sync", str(lockfile_path)]
        for extra in self.features:
            command.extend(["--extra", extra])
        for group in self.dependency_groups:
            command.extend(["--group", group])
        add_verbosity_flag(command, self.verbosity, adjustment=-1)
        if dry_run:
            command.append("--dry-run")
        python_version = self.config.get("python", "")
        if python_version:
            command.extend(["--python-version", python_version])
        return command

    def dependencies_in_sync(self):
        with self.safe_activation():
            workspace_deps = [dep for dep in self.local_dependencies_complex if dep.path]
            if self.distributions.missing_dependencies(workspace_deps):
                return False

            if self.locked:
                from hatch.env.lock import (
                    LockerNotFoundError,
                    LockerUnsupportedError,
                    get_locker_plugin_class,
                    resolve_lockfile_path,
                )

                lockfile_path = resolve_lockfile_path(self)
                if lockfile_path.is_file():
                    try:
                        locker_cls = get_locker_plugin_class(self.app.project, self)
                    except (LockerNotFoundError, LockerUnsupportedError):
                        return False
                    if not locker_cls.install_matches_lock(self, lockfile_path):
                        return False

            return not self.missing_dependencies

    def sync_dependencies(self):
        with self.safe_activation():
            workspace_deps = [dep for dep in self.local_dependencies_complex if dep.path]
            workspace_names = {dep.name.lower() for dep in self.local_dependencies_complex if dep.path}

            if self.locked:
                from hatch.env.lock import apply_lock_with_locker, resolve_lockfile_path

                lockfile_path = resolve_lockfile_path(self)
                if lockfile_path.is_file():
                    all_install_args = []
                    for dep in workspace_deps:
                        if dep.editable:
                            all_install_args.extend(["--editable", dep.path])
                        else:
                            all_install_args.append(dep.path)

                    if all_install_args:
                        self.platform.check_command(self.construct_pip_install_command(all_install_args))
                    apply_lock_with_locker(self, lockfile_path)
                    return

            # If we do not have missing dependencies we should not sync
            if not self.missing_dependencies:
                return

            all_install_args = []

            for dep in workspace_deps:
                if dep.editable:
                    all_install_args.extend(["--editable", dep.path])
                else:
                    all_install_args.append(dep.path)

            standard_dependencies = []

            user_editable_dependencies = []

            for dependency in self.missing_dependencies:
                # Skip if already handled as workspace dependency
                if dependency.name.lower() in workspace_names:
                    continue

                if dependency.editable and dependency.path is not None:
                    user_editable_dependencies.append(str(dependency.path))
                else:
                    standard_dependencies.append(str(dependency))

            all_install_args.extend(standard_dependencies)

            for dep_path in user_editable_dependencies:
                all_install_args.extend(["--editable", dep_path])

            if all_install_args:
                self.platform.check_command(self.construct_pip_install_command(all_install_args))

    @contextmanager
    def command_context(self):
        with self.safe_activation():
            yield

    def construct_pip_install_command(self, args: list[str]):
        if not self.use_uv:
            return super().construct_pip_install_command(args)

        command = [self.uv_path, "pip", "install"]

        # Default to -1 verbosity
        add_verbosity_flag(command, self.verbosity, adjustment=-1)

        command.extend(args)
        return command

    def enter_shell(self, name: str, path: str, args: Iterable[str]):
        shell_executor = getattr(self.shells, f"enter_{name}", None)
        if shell_executor is None:
            # Manually activate in lieu of an activation script
            with self.safe_activation():
                self.platform.exit_with_command([path, *args])
        else:
            with self.expose_uv(), self.get_env_vars():
                shell_executor(path, args, self.virtual_env.executables_directory)

    def check_compatibility(self):
        super().check_compatibility()

        python_version = self.config.get("python", "")
        if (
            os.environ.get(AppEnvVars.PYTHON)
            or self._find_existing_interpreter(python_version) is not None
            or self._get_available_distribution(python_version) is not None
        ):
            return

        message = (
            f"cannot locate Python: {python_version}"
            if python_version
            else "no compatible Python distribution available"
        )
        raise OSError(message)

    @cached_property
    def _preferred_python_version(self):
        version = f"{sys.version_info.major}.{sys.version_info.minor}"

        if FREETHREADED_BUILD:
            version += "t"

        return version

    @cached_property
    def parent_python(self):
        if python_choice := self.config.get("python", ""):
            return self._get_concrete_interpreter_path(python_choice)

        if explicit_default := os.environ.get(AppEnvVars.PYTHON):
            return sys.executable if explicit_default == "self" else explicit_default

        return self._get_concrete_interpreter_path()

    @cached_property
    def python_manager(self) -> PythonManager:
        from hatch.python.core import PythonManager

        return PythonManager(self.isolated_data_directory / ".pythons")

    def get_interpreter_resolver_env(self) -> dict[str, str]:
        env = dict(os.environ)
        python_dirs = [str(dist.python_path.parent) for dist in self.python_manager.get_installed().values()]
        if not python_dirs:
            return env

        internal_path = os.pathsep.join(python_dirs)
        old_path = env.pop("PATH", None)
        env["PATH"] = internal_path if old_path is None else f"{old_path}{os.pathsep}{internal_path}"

        return env

    def upgrade_possible_internal_python(self, python_path: str) -> None:
        if "internal" not in self._python_sources:
            return

        for dist in self.python_manager.get_installed().values():
            if dist.python_path == Path(python_path):
                if dist.needs_update():
                    with self.app.status(f"Updating Python distribution: {dist.name}"):
                        self.python_manager.install(dist.name)

                break

    def _interpreter_is_compatible(self, interpreter: PythonInfo) -> bool:
        return (
            interpreter.executable is not None
            and self._is_stable_path(interpreter.executable)
            and (self.skip_install or self._python_constraint.contains(interpreter.version_str))
        )

    def _get_concrete_interpreter_path(self, python_version: str = "") -> str | None:
        known_resolvers = self._python_resolvers()
        resolvers = [known_resolvers[source] for source in self._python_sources]
        if python_version:
            for resolver in resolvers:
                if (concrete_path := resolver(python_version)) is not None:
                    return concrete_path
        else:
            # Prefer the Python version Hatch is currently using
            for resolver in resolvers:
                if (concrete_path := resolver(self._preferred_python_version)) is not None:
                    return concrete_path

            # Fallback to whatever is compatible
            for resolver in resolvers:
                if (concrete_path := resolver("")) is not None:
                    return concrete_path

        return None

    def _resolve_external_interpreter_path(self, python_version: str) -> str | None:
        if (existing_path := self._find_existing_interpreter(python_version)) is not None:
            self.upgrade_possible_internal_python(existing_path)
            return existing_path

        return None

    def _resolve_internal_interpreter_path(self, python_version: str) -> str | None:
        if (available_distribution := self._get_available_distribution(python_version)) is not None:
            with self.app.status(f"Installing Python distribution: {available_distribution}"):
                dist = self.python_manager.install(available_distribution)

            return str(dist.python_path)

        return None

    def _find_existing_interpreter(self, python_version: str = "") -> str | None:
        import python_discovery

        python_info = python_discovery.get_interpreter(
            python_version, (), env=self.get_interpreter_resolver_env(), predicate=self._interpreter_is_compatible
        )
        return None if python_info is None else python_info.executable

    def _get_available_distribution(self, python_version: str = "") -> str | None:
        from hatch.python.resolve import get_compatible_distributions

        compatible_distributions = get_compatible_distributions()
        for installed_distribution in self.python_manager.get_installed():
            compatible_distributions.pop(installed_distribution, None)

        if not python_version:
            # Only try providing CPython distributions
            available_distributions = [d for d in compatible_distributions if not d.startswith("pypy")]

            # Prioritize the version that Hatch is currently using, if available
            with suppress(ValueError):
                available_distributions.remove(self._preferred_python_version)
                available_distributions.append(self._preferred_python_version)

            # Latest first
            available_distributions.reverse()
        elif python_version in compatible_distributions:
            available_distributions = [python_version]
        else:
            return None

        for available_distribution in available_distributions:
            minor_version = (
                available_distribution.replace("pypy", "", 1)
                if available_distribution.startswith("pypy")
                else available_distribution
            )
            if not self._python_constraint.contains(minor_version):
                continue

            return available_distribution

        return None

    def _is_stable_path(self, executable: str) -> bool:
        path = Path(executable).resolve()
        parents = path.parents

        # https://pypa.github.io/pipx/how-pipx-works/
        if (Path.home() / ".local" / "pipx" / "venvs") in parents:
            return False

        from platformdirs import user_data_dir

        # https://github.com/ofek/pyapp/blob/v0.13.0/src/app.rs#L27
        if Path(user_data_dir("pyapp", appauthor=False)) in parents:
            return False

        # via Windows store
        if self.platform.windows and str(path).endswith("WindowsApps\\python.exe"):
            return False

        # via Homebrew
        return not (self.platform.macos and Path("/usr/local/Cellar") in parents)

    @cached_property
    def _python_sources(self) -> list[str]:
        return self.config.get("python-sources") or ["external", "internal"]

    def _python_resolvers(self) -> dict[str, Callable[[str], str | None]]:
        return {
            "external": self._resolve_external_interpreter_path,
            "internal": self._resolve_internal_interpreter_path,
        }

    @cached_property
    def _python_constraint(self) -> SpecifierSet:
        from packaging.specifiers import SpecifierSet

        # Note that we do not support this field being dynamic because if we were to set up the
        # build environment to retrieve the field then we would be stuck because we need to use
        # a satisfactory version to set up the environment
        return SpecifierSet(self.metadata.config.get("project", {}).get("requires-python", ""))

    @contextmanager
    def safe_activation(self):
        # In order of precedence:
        # - This environment
        # - UV
        # - User-defined environment variables
        with self.get_env_vars(), self.expose_uv(), self:
            yield



from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING

from hatch.env.lockers.interface import LockerInterface
from hatch.env.utils import add_verbosity_flag
from hatch.utils.fs import Path

if TYPE_CHECKING:
    from hatch.env.plugin.interface import EnvironmentInterface
    from hatch.env.virtual import VirtualEnvironment


class UvLocker(LockerInterface):
    PLUGIN_NAME = "uv"

    @classmethod
    def supports(cls, environment: EnvironmentInterface) -> bool:
        from hatch.env.virtual import VirtualEnvironment

        return isinstance(environment, VirtualEnvironment) and environment.use_uv

    @classmethod
    def generate(
        cls,
        environment: EnvironmentInterface,
        dependencies: list[str],
        output_path: Path,
        *,
        upgrade: bool = False,
        upgrade_packages: tuple[str, ...] = (),
        layered: bool = False,
        lock_extras: tuple[str, ...] = (),
        lock_groups: tuple[str, ...] = (),
    ) -> None:
        from hatch.env.virtual import VirtualEnvironment

        if not isinstance(environment, VirtualEnvironment):
            message = "UvLocker requires a virtual environment with UV"
            raise TypeError(message)

        env = environment
        deps_file: Path | None = None
        if dependencies:
            with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False, encoding="utf-8") as f:
                f.write("\n".join(dependencies))
                f.write("\n")
                deps_file = Path(f.name)

        try:
            cls._compile(
                env,
                output_path,
                upgrade=upgrade,
                upgrade_packages=upgrade_packages,
                layered=layered,
                lock_extras=lock_extras,
                lock_groups=lock_groups,
                requirements_file=deps_file,
            )
        finally:
            if deps_file is not None:
                deps_file.unlink(missing_ok=True)

    @classmethod
    def _compile(
        cls,
        environment: VirtualEnvironment,
        output_path: Path,
        *,
        upgrade: bool,
        upgrade_packages: tuple[str, ...],
        layered: bool,
        lock_extras: tuple[str, ...],
        lock_groups: tuple[str, ...],
        requirements_file: Path | None,
    ) -> None:
        pyproject = environment.root / "pyproject.toml"
        command = [environment.uv_path, "pip", "compile"]

        if requirements_file is not None:
            command.append(str(requirements_file))
        if layered:
            command.append(str(pyproject))

        command.extend([
            "--generate-hashes",
            "--output-file",
            str(output_path),
        ])

        for extra in lock_extras:
            command.extend(["--extra", extra])
        for group in lock_groups:
            command.extend(["--group", group])

        add_verbosity_flag(command, environment.verbosity, adjustment=-1)

        if upgrade:
            command.append("--upgrade")
        for pkg in upgrade_packages:
            command.extend(["--upgrade-package", pkg])

        python_version = environment.config.get("python", "")
        if python_version:
            command.extend(["--python-version", python_version])

        with environment.command_context():
            environment.platform.check_command(command)

    @classmethod
    def in_sync(
        cls,
        environment: EnvironmentInterface,
        dependencies: list[str],
        output_path: Path,
        *,
        upgrade: bool = False,
        upgrade_packages: tuple[str, ...] = (),
        layered: bool = False,
        lock_extras: tuple[str, ...] = (),
        lock_groups: tuple[str, ...] = (),
    ) -> bool:
        if not output_path.is_file():
            return False

        with tempfile.NamedTemporaryFile(mode="w", suffix=".toml", delete=False, encoding="utf-8") as f:
            temp_path = Path(f.name)

        try:
            cls.generate(
                environment,
                dependencies,
                temp_path,
                upgrade=upgrade,
                upgrade_packages=upgrade_packages,
                layered=layered,
                lock_extras=lock_extras,
                lock_groups=lock_groups,
            )
            existing = output_path.read_text(encoding="utf-8")
            fresh = temp_path.read_text(encoding="utf-8")
        finally:
            temp_path.unlink(missing_ok=True)

        return existing == fresh

    @classmethod
    def apply_lock(cls, environment: EnvironmentInterface, lock_path: Path) -> None:
        from hatch.env.virtual import VirtualEnvironment

        if not isinstance(environment, VirtualEnvironment):
            message = "UvLocker.apply_lock requires a virtual environment"
            raise TypeError(message)

        with environment.command_context():
            environment.platform.check_command(environment.uv_pip_sync_command(lock_path))

    @classmethod
    def install_matches_lock(cls, environment: EnvironmentInterface, lock_path: Path) -> bool:
        from hatch.env.virtual import VirtualEnvironment

        if not isinstance(environment, VirtualEnvironment):
            return True

        completed = environment.platform.run_command(
            environment.uv_pip_sync_command(lock_path, dry_run=True),
            capture_output=True,
        )
        return completed.returncode == 0



from __future__ import annotations

import json
import os

from hatchling.bridge.app import Application
from hatchling.metadata.core import ProjectMetadata
from hatchling.plugin.manager import PluginManager

RUNNER: dict = {}


def main() -> None:
    project_root: str = RUNNER["project_root"]
    output_dir: str = RUNNER["output_dir"]
    targets: list[str] = RUNNER["targets"]

    app = Application()
    plugin_manager = PluginManager()
    metadata = ProjectMetadata(project_root, plugin_manager)

    dependencies: dict[str, None] = {}
    for target_name in targets:
        builder_class = plugin_manager.builder.get(target_name)
        if builder_class is None:
            continue

        builder = builder_class(
            project_root, plugin_manager=plugin_manager, metadata=metadata, app=app.get_safe_application()
        )
        for dependency in builder.config.dependencies:
            dependencies[dependency] = None

    output = json.dumps(list(dependencies))
    with open(os.path.join(output_dir, "output.json"), "w", encoding="utf-8") as f:
        f.write(output)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Show the available Python distributions")
@click.option("--ascii", "force_ascii", is_flag=True, help="Whether or not to only use ASCII characters")
@click.option("--dir", "-d", "directory", help="The directory in which distributions reside")
@click.pass_obj
def show(app: Application, *, force_ascii: bool, directory: str | None):
    """Show the available Python distributions."""
    from hatch.python.resolve import get_compatible_distributions

    manager = app.get_python_manager(directory)
    installed = manager.get_installed()

    installed_columns: dict[str, dict[int, str]] = {"Name": {}, "Version": {}, "Status": {}}
    for i, (name, installed_dist) in enumerate(installed.items()):
        installed_columns["Name"][i] = name
        installed_columns["Version"][i] = installed_dist.version
        if installed_dist.needs_update():
            installed_columns["Status"][i] = "Update available"

    available_columns: dict[str, dict[int, str]] = {"Name": {}, "Version": {}}
    for i, (name, dist) in enumerate(get_compatible_distributions().items()):
        if name in installed:
            continue

        available_columns["Name"][i] = name
        available_columns["Version"][i] = dist.version.base_version

    app.display_table("Installed", installed_columns, show_lines=True, force_ascii=force_ascii)
    app.display_table("Available", available_columns, show_lines=True, force_ascii=force_ascii)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from hatchling.metadata.utils import get_normalized_dependency, normalize_project_name

if TYPE_CHECKING:
    from packaging.requirements import Requirement

    from hatch.dep.core import Dependency


def normalize_marker_quoting(text: str) -> str:
    # All TOML writers use double quotes, so allow copy/pasting to avoid escaping
    return text.replace('"', "'")


def get_normalized_dependencies(requirements: list[Requirement]) -> list[str]:
    normalized_dependencies = {get_normalized_dependency(requirement) for requirement in requirements}
    return sorted(normalized_dependencies)


def hash_dependencies(requirements: list[Dependency]) -> str:
    from hashlib import sha256

    data = "".join(
        sorted(
            # Internal spacing is ignored by PEP 440
            normalized_dependency.replace(" ", "")
            for normalized_dependency in {get_normalized_dependency(req) for req in requirements}
        )
    ).encode("utf-8")

    return sha256(data).hexdigest()


def get_complex_dependencies(dependencies: list[str]) -> dict[str, Dependency]:
    from hatch.dep.core import Dependency

    dependencies_complex = {}
    for dependency in dependencies:
        dependencies_complex[dependency] = Dependency(dependency)

    return dependencies_complex


def get_complex_features(features: dict[str, list[str]]) -> dict[str, dict[str, Dependency]]:
    from hatch.dep.core import Dependency

    optional_dependencies_complex = {}
    for feature, optional_dependencies in features.items():
        optional_dependencies_complex[feature] = {
            optional_dependency: Dependency(optional_dependency) for optional_dependency in optional_dependencies
        }

    return optional_dependencies_complex


def get_complex_dependency_group(
    dependency_groups: dict[str, Any], group: str, past_groups: tuple[str, ...] = ()
) -> list[Dependency]:
    from hatch.dep.core import Dependency

    if group in past_groups:
        msg = f"Cyclic dependency group include: {group} -> {past_groups}"
        raise ValueError(msg)

    if group not in dependency_groups:
        msg = f"Dependency group '{group}' not found"
        raise LookupError(msg)

    raw_group = dependency_groups[group]
    if not isinstance(raw_group, list):
        msg = f"Dependency group '{group}' is not a list"
        raise TypeError(msg)

    realized_group = []
    for item in raw_group:
        if isinstance(item, str):
            realized_group.append(Dependency(item))
        elif isinstance(item, dict):
            if tuple(item.keys()) != ("include-group",):
                msg = f"Invalid dependency group item: {item}"
                raise ValueError(msg)

            include_group = normalize_project_name(next(iter(item.values())))
            realized_group.extend(get_complex_dependency_group(dependency_groups, include_group, (*past_groups, group)))
        else:
            msg = f"Invalid dependency group item: {item}"
            raise TypeError(msg)

    return realized_group



from __future__ import annotations

from typing import TYPE_CHECKING

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


@click.command(short_help="Remove all environments")
@click.pass_obj
def prune(app: Application):
    """Remove all environments."""
    app.ensure_environment_plugin_dependencies()

    environment_types = app.plugins.environment.collect()

    for environments in (app.project.config.envs, app.project.config.internal_envs):
        for env_name in environments:
            if env_name == app.env_active:
                app.abort(f"Cannot remove active environment: {env_name}")

        for env_name, config in environments.items():
            environment_type = config["type"]
            if environment_type not in environment_types:
                app.abort(f"Environment `{env_name}` has unknown type: {environment_type}")

            environment = environment_types[environment_type](
                app.project.location,
                app.project.metadata,
                env_name,
                config,
                app.project.config.matrix_variables.get(env_name, {}),
                app.get_env_directory(environment_type),
                app.data_dir / "env" / environment_type,
                app.platform,
                app.verbosity,
                app,
            )
            if environment.exists():
                with app.status(f"Removing environment: {env_name}"):
                    environment.remove()



from __future__ import annotations

from typing import Any


def get_default_config() -> dict[str, Any]:
    from hatch.env.internal.test import get_default_config as get_test_config

    test_config = get_test_config()
    test_deps = test_config.get("dependencies", [])

    return {
        "installer": "uv",
        "dependencies": [f"pyrefly=={PYREFLY_DEFAULT_VERSION}", *test_deps],
        "scripts": {
            "check": "pyrefly check{env:HATCH_CHECK_TYPES_ARGS:} {args}",
            "check-summarize": "pyrefly check{env:HATCH_CHECK_TYPES_ARGS:} --summarize-errors {args}",
            "coverage": "pyrefly report{env:HATCH_CHECK_TYPES_ARGS:} {args}",
            "config": "pyrefly dump-config {args}",
        },
    }


PYREFLY_DEFAULT_VERSION: str = "1.0.0"



import click

from hatch.cli.python.find import find
from hatch.cli.python.install import install
from hatch.cli.python.remove import remove
from hatch.cli.python.show import show
from hatch.cli.python.update import update


@click.group(short_help="Manage Python installations")
def python():
    pass


python.add_command(find)
python.add_command(install)
python.add_command(remove)
python.add_command(show)
python.add_command(update)



from hatch.env.system import SystemEnvironment
from hatch.env.virtual import VirtualEnvironment
from hatchling.plugin import hookimpl


@hookimpl
def hatch_register_environment():
    return [SystemEnvironment, VirtualEnvironment]



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import click

if TYPE_CHECKING:
    from hatch.cli.application import Application


def get_install_source(platform_name: str) -> str:
    import os
    import sys
    import sysconfig

    from platformdirs import user_data_dir

    from hatch.utils.fs import Path

    default_source = "pip"

    python_path = Path(sys.executable).resolve()
    parent_paths = python_path.parents

    # https://github.com/ofek/pyapp/blob/v0.13.0/src/app.rs#L27
    if Path(user_data_dir("pyapp", appauthor=False)) in parent_paths:
        return "binary"

    # https://pypa.github.io/pipx/how-pipx-works/
    if (Path.home() / ".local" / "pipx" / "venvs") in parent_paths:
        return "pipx"

    # https://packaging.python.org/en/latest/specifications/externally-managed-environments/#marking-an-interpreter-as-using-an-external-package-manager
    try:
        stdlib_path_config = sysconfig.get_path("stdlib")
    # https://docs.python.org/3/library/sysconfig.html#sysconfig.get_path
    except KeyError:
        stdlib_path_config = ""

    if (
        # This does not work on NixOS, see: https://github.com/NixOS/nixpkgs/issues/201037
        sys.prefix == sys.base_prefix
        and stdlib_path_config
        and (stdlib_path := Path(stdlib_path_config)).is_dir()
        and any(p.name == "EXTERNALLY-MANAGED" and p.is_file() for p in stdlib_path.iterdir())
    ):
        return "system"

    if platform_name == "windows":
        if sys.executable.endswith("WindowsApps\\python.exe"):
            return "Windows Store"

        # Break early because nothing after is applicable
        return default_source

    if platform_name == "macos" and Path("/usr/local/Cellar") in parent_paths:  # no cov
        return "Homebrew"

    # https://github.com/pyenv/pyenv/tree/v2.3.35#set-up-your-shell-environment-for-pyenv
    if Path(os.environ.get("PYENV_ROOT", "~/.pyenv")).expand() in parent_paths:
        return "Pyenv"

    return default_source


@click.command(short_help="Generate a pre-populated GitHub issue")
@click.option("--no-open", "-n", is_flag=True, help="Show the URL instead of opening it")
@click.pass_obj
def report(app: Application, *, no_open: bool) -> None:
    """Generate a pre-populated GitHub issue."""
    import sys
    import webbrowser
    from textwrap import indent
    from urllib.parse import quote_plus

    import tomlkit

    from hatch._version import __version__
    from hatch.utils.toml import load_toml_data

    # Retain the config that would be most useful
    full_config = load_toml_data(app.config_file.read_scrubbed())
    relevant_config: dict[str, Any] = {}
    for setting in ("mode", "shell"):
        if setting in full_config:
            relevant_config[setting] = full_config[setting]

    full_dirs = full_config.get("dirs")
    if isinstance(full_dirs, dict) and (env_dirs := full_dirs.get("envs")):
        relevant_config["dirs"] = {"envs": env_dirs}

    # Try to determine how Hatch was installed
    source = get_install_source(app.platform.name)

    element_padding = " " * 4
    body = f"""\
## Current behavior
<!-- A clear and concise description of the behavior. -->

## Expected behavior
<!-- A clear and concise description of what you expected to happen. -->

## Additional context
<!-- Add any other context about the problem here. If applicable, add screenshots to help explain. -->

## Debug

### Installation

- Source: {source}
- Version: {__version__}
- Platform: {app.platform.display_name}
- Python version:
{element_padding}```
{indent(sys.version, element_padding)}
{element_padding}```

### Configuration

```toml
{tomlkit.dumps(relevant_config).rstrip()}
```
"""

    url = f"https://github.com/pypa/hatch/issues/new?body={quote_plus(body)}"
    if no_open:
        app.display(url)
    else:
        webbrowser.open_new_tab(url)



import click

from hatch.cli.env.lock import dependency_lock_click_options, run_dep_lock


@click.group(short_help="Manage environment dependencies")
def dep():
    pass


@dep.command("lock", short_help="Generate a lockfile for the active environment (`-e` / `HATCH_ENV`)")
@dependency_lock_click_options
@click.pass_obj
def dep_lock(
    app,
    *,
    upgrade: bool,
    upgrade_package: tuple[str, ...],
    export_path: str | None,
    export_all_path: str | None,
    check: bool,
):
    """Resolve dependencies and write a PEP 751 ``pylock.toml`` for the selected environment."""
    app.ensure_environment_plugin_dependencies()
    run_dep_lock(
        app,
        upgrade=upgrade,
        upgrade_package=upgrade_package,
        export_path=export_path,
        export_all_path=export_all_path,
        check=check,
    )


@dep.command("sync", short_help="Install dependencies from the environment lockfile (`apply_lock`)")
@click.pass_obj
def dep_sync(app):
    """Sync the active environment to its lockfile (``locked`` environments only)."""
    app.ensure_environment_plugin_dependencies()

    environment = app.project.get_environment()
    if not environment.locked:
        app.abort(
            "The active environment is not `locked`. Set `locked = true` or use `hatch env lock --export` "
            "and a normal install workflow."
        )

    from hatch.env.lock import LockerNotFoundError, LockerUnsupportedError, resolve_lockfile_path

    lock_path = resolve_lockfile_path(environment)
    if not lock_path.is_file():
        app.abort(f"No lockfile at `{lock_path}`. Run `hatch dep lock` or `hatch env lock` first.")

    try:
        with app.status("Syncing from lockfile"):
            environment.sync_dependencies()
    except LockerNotFoundError as e:
        app.abort(str(e))
    except LockerUnsupportedError as e:
        if e.detail:
            app.abort(f"Cannot sync environment `{environment.name}` from lockfile: {e.detail}")
        app.abort(str(e))

    app.display_success(f"Synced environment `{environment.name}` from `{lock_path.name}`")


@dep.command("hash", short_help="Output a hash of the currently defined dependencies")
@click.option("--project-only", "-p", is_flag=True, help="Whether or not to exclude environment dependencies")
@click.option("--env-only", "-e", is_flag=True, help="Whether or not to exclude project dependencies")
@click.pass_obj
def hash_dependencies(app, project_only, env_only):
    """Output a hash of the currently defined dependencies."""
    app.ensure_environment_plugin_dependencies()

    from hatch.utils.dep import get_complex_dependencies, hash_dependencies

    environment = app.project.get_environment()

    all_requirements = []
    if project_only:
        dependencies, _ = app.project.get_dependencies()
        dependencies_complex = get_complex_dependencies(dependencies)
        all_requirements.extend(dependencies_complex.values())
    elif env_only:
        all_requirements.extend(environment.environment_dependencies_complex)
    else:
        dependencies, _ = app.project.get_dependencies()
        dependencies_complex = get_complex_dependencies(dependencies)
        all_requirements.extend(dependencies_complex.values())
        all_requirements.extend(environment.environment_dependencies_complex)

    app.display(hash_dependencies(all_requirements))


@dep.group(short_help="Display dependencies in various formats")
def show():
    pass


@show.command(short_help="Enumerate dependencies in a tabular format")
@click.option("--project-only", "-p", is_flag=True, help="Whether or not to exclude environment dependencies")
@click.option("--env-only", "-e", is_flag=True, help="Whether or not to exclude project dependencies")
@click.option("--lines", "-l", "show_lines", is_flag=True, help="Whether or not to show lines between table rows")
@click.option("--ascii", "force_ascii", is_flag=True, help="Whether or not to only use ASCII characters")
@click.pass_obj
def table(app, project_only, env_only, show_lines, force_ascii):
    """Enumerate dependencies in a tabular format."""
    app.ensure_environment_plugin_dependencies()

    from hatch.dep.core import Dependency
    from hatch.utils.dep import get_complex_dependencies, get_normalized_dependencies, normalize_marker_quoting

    environment = app.project.get_environment()

    project_requirements = []
    environment_requirements = []
    if project_only:
        dependencies, _ = app.project.get_dependencies()
        dependencies_complex = get_complex_dependencies(dependencies)
        project_requirements.extend(dependencies_complex.values())
    elif env_only:
        environment_requirements.extend(environment.environment_dependencies_complex)
    else:
        dependencies, _ = app.project.get_dependencies()
        dependencies_complex = get_complex_dependencies(dependencies)
        project_requirements.extend(dependencies_complex.values())
        environment_requirements.extend(environment.environment_dependencies_complex)

    for all_requirements, table_title in (
        (project_requirements, "Project"),
        (environment_requirements, f"Env: {app.env}"),
    ):
        if not all_requirements:
            continue

        normalized_requirements = [Dependency(d) for d in get_normalized_dependencies(all_requirements)]

        columns = {"Name": {}, "URL": {}, "Versions": {}, "Markers": {}, "Features": {}}
        for i, requirement in enumerate(normalized_requirements):
            columns["Name"][i] = requirement.name

            if requirement.url:
                columns["URL"][i] = str(requirement.url)

            if requirement.specifier:
                columns["Versions"][i] = str(requirement.specifier)

            if requirement.marker:
                columns["Markers"][i] = normalize_marker_quoting(str(requirement.marker))

            if requirement.extras:
                columns["Features"][i] = ", ".join(sorted(requirement.extras))

        column_options = {}
        for column_title in columns:
            if column_title != "URL":
                column_options[column_title] = {"no_wrap": True}

        app.display_table(
            table_title, columns, show_lines=show_lines, column_options=column_options, force_ascii=force_ascii
        )


@show.command(short_help="Enumerate dependencies as a list of requirements")
@click.option("--project-only", "-p", is_flag=True, help="Whether or not to exclude environment dependencies")
@click.option("--env-only", "-e", is_flag=True, help="Whether or not to exclude project dependencies")
@click.option(
    "--feature",
    "-f",
    "features",
    multiple=True,
    help="Whether or not to only show the dependencies of the specified features",
)
@click.option("--all", "all_features", is_flag=True, help="Whether or not to include the dependencies of all features")
@click.pass_obj
def requirements(app, project_only, env_only, features, all_features):
    """Enumerate dependencies as a list of requirements."""
    app.ensure_environment_plugin_dependencies()

    from hatch.utils.dep import get_complex_dependencies, get_complex_features, get_normalized_dependencies
    from hatchling.metadata.utils import normalize_project_name

    environment = app.project.get_environment()
    dependencies, optional_dependencies = app.project.get_dependencies()
    dependencies_complex = get_complex_dependencies(dependencies)
    optional_dependencies_complex = get_complex_features(optional_dependencies)

    all_requirements = []
    if features:
        for raw_feature in features:
            feature = normalize_project_name(raw_feature)
            if feature not in optional_dependencies_complex:
                app.abort(f"Feature `{feature}` is not defined in field `project.optional-dependencies`")

            all_requirements.extend(optional_dependencies_complex[feature].values())
    elif project_only:
        all_requirements.extend(dependencies_complex.values())
    elif env_only:
        all_requirements.extend(environment.environment_dependencies_complex)
    else:
        all_requirements.extend(dependencies_complex.values())
        all_requirements.extend(environment.environment_dependencies_complex)

    if not features and all_features:
        for optional_dependencies in optional_dependencies_complex.values():
            all_requirements.extend(optional_dependencies.values())

    for dependency in get_normalized_dependencies(all_requirements):
        app.display(dependency)



from hatch.template import File
from hatch.utils.fs import Path


class PackageRoot(File):
    def __init__(
        self,
        template_config: dict,
        plugin_config: dict,  # noqa: ARG002
    ):
        super().__init__(Path(template_config["package_name"], "__init__.py"), "")


class MetadataFile(File):
    def __init__(
        self,
        template_config: dict,
        plugin_config: dict,  # noqa: ARG002
    ):
        super().__init__(Path(template_config["package_name"], "__about__.py"), '__version__ = "0.0.1"\n')


class Readme(File):
    TEMPLATE = """\
# {project_name}

[![PyPI - Version](https://img.shields.io/pypi/v/{project_name_normalized}.svg)](https://pypi.org/project/{project_name_normalized})
[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/{project_name_normalized}.svg)](https://pypi.org/project/{project_name_normalized})
{extra_badges}
-----

## Table of Contents

- [Installation](#installation)
{extra_toc}
## Installation

```console
pip install {project_name_normalized}
```{license_info}
"""

    def __init__(
        self,
        template_config: dict,
        plugin_config: dict,  # noqa: ARG002
    ):
        extra_badges = ""
        extra_toc = ""

        license_info = ""
        if template_config["license_data"]:
            extra_toc += "- [License](#license)\n"
            license_info += (
                f"\n\n## License\n\n`{template_config['project_name_normalized']}` is distributed under the terms of "
            )

            license_data = template_config["license_data"]
            if len(license_data) == 1:
                license_id = next(iter(license_data))
                license_info += f"the [{license_id}](https://spdx.org/licenses/{license_id}.html) license."
            else:
                license_info += "any of the following licenses:\n"
                for license_id in sorted(license_data):
                    license_info += f"\n- [{license_id}](https://spdx.org/licenses/{license_id}.html)"

        super().__init__(
            Path(template_config["readme_file_path"]),
            self.TEMPLATE.format(
                extra_badges=extra_badges, extra_toc=extra_toc, license_info=license_info, **template_config
            ),
        )


class PyProject(File):
    TEMPLATE = """\
[build-system]
requires = ["hatchling"]
build-backend = "hatchling.build"

[project]
name = "{project_name_normalized}"
dynamic = ["version"]
description = {description!r}
readme = "{readme_file_path}"
requires-python = ">=3.8"
license = "{license_expression}"{license_files}
keywords = []
authors = [
  {{ name = "{name}", email = "{email}" }},
]
classifiers = [
  "Development Status :: 4 - Beta",
  "Programming Language :: Python",
  "Programming Language :: Python :: 3.8",
  "Programming Language :: Python :: 3.9",
  "Programming Language :: Python :: 3.10",
  "Programming Language :: Python :: 3.11",
  "Programming Language :: Python :: 3.12",
  "Programming Language :: Python :: Implementation :: CPython",
  "Programming Language :: Python :: Implementation :: PyPy",
]
dependencies = {dependency_data}

[project.urls]{project_url_data}{cli_scripts}

[tool.hatch.version]
path = "{package_metadata_file_path}"{tests_section}
"""

    def __init__(self, template_config: dict, plugin_config: dict):
        template_config = dict(template_config)
        template_config["name"] = repr(template_config["name"])[1:-1]

        project_url_data = ""
        project_urls = (
            plugin_config["project_urls"]
            if "project_urls" in plugin_config
            else {
                "Documentation": "https://github.com/{name}/{project_name_normalized}#readme",
                "Issues": "https://github.com/{name}/{project_name_normalized}/issues",
                "Source": "https://github.com/{name}/{project_name_normalized}",
            }
        )
        if project_urls:
            for label, url in project_urls.items():
                normalized_label = f'"{label}"' if " " in label else label
                project_url_data += f'\n{normalized_label} = "{url.format(**template_config)}"'

        dependency_data = "["
        if template_config["dependencies"]:
            for dependency in sorted(template_config["dependencies"]):
                dependency_data += f'\n  "{dependency}",\n'
        dependency_data += "]"

        cli_scripts = ""
        if template_config["args"]["cli"]:
            cli_scripts = f"""

[project.scripts]
{template_config["project_name_normalized"]} = "{template_config["package_name"]}.cli:{template_config["package_name"]}"\
"""

        tests_section = ""
        if plugin_config["tests"]:
            package_location = "src/" if plugin_config["src-layout"] else ""
            tests_section = f"""

[tool.hatch.envs.types]
extra-dependencies = [
  "mypy>=1.0.0",
]
[tool.hatch.envs.types.scripts]
check = "mypy --install-types --non-interactive {{args:{package_location}{template_config["package_name"]} tests}}"

[tool.coverage.run]
source_pkgs = ["{template_config["package_name"]}", "tests"]
branch = true
parallel = true
omit = [
  "{package_location}{template_config["package_name"]}/__about__.py",
]

[tool.coverage.paths]
{template_config["package_name"]} = ["{package_location}{template_config["package_name"]}", "*/{template_config["project_name_normalized"]}/{package_location}{template_config["package_name"]}"]
tests = ["tests", "*/{template_config["project_name_normalized"]}/tests"]

[tool.coverage.report]
exclude_lines = [
  "no cov",
  "if __name__ == .__main__.:",
  "if TYPE_CHECKING:",
]"""

        super().__init__(
            Path("pyproject.toml"),
            self.TEMPLATE.format(
                project_url_data=project_url_data,
                dependency_data=dependency_data,
                cli_scripts=cli_scripts,
                tests_section=tests_section,
                **template_config,
            ),
        )
